import re
from enum import Enum
from typing import Any, List, Optional
from urllib.parse import quote

import pandas
from pandas import DataFrame

from .. import mdapi
from .._base import _logger
from ..direct._config import _Config
from . import _decorator, _error_messages
from ._base_api import APIBackend
from ._config_key import ALL_ASSET_FLOW_DATA_POINTS
from ._data_point import (
    _get_all_universes,
    _get_asset_flow_data_points_by_ids,
    _get_data_point_details,
)
from ._exceptions import BadRequestException, ResourceNotFoundError, ValueErrorException
from ._utils import (
    _empty_to_none,
    _extract_data,
    _filter_data_frame_column_by_setting,
    _reindex_data_frame_column,
)
from .data_type import InvestmentIdentifier, Universe

_config = _Config()


class _LookupModule(Enum):
    DATA_SET = 0
    DATA_POINT = 1
    INVESTMENT = 2


class LookupDataAPIBackend(APIBackend):
    """
    Subclass to call the Lookup Data API and handle any HTTP errors that occur.
    """

    def __init__(self, lookup_module: _LookupModule) -> None:
        super().__init__()
        self._lookup_module = lookup_module

    def _get_resource_not_found_message(self) -> str:
        if self._lookup_module == _LookupModule.DATA_SET:
            return _error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_SET_LOOKUP
        elif self._lookup_module == _LookupModule.DATA_POINT:
            return _error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_POINT_LOOKUP
        return _error_messages.RESOURCE_NOT_FOUND_ERROR

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json().get("message")
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(self._get_resource_not_found_message()) from None


_data_set_lookup_api_request = LookupDataAPIBackend(_LookupModule.DATA_SET)
_investment_lookup_api_request = LookupDataAPIBackend(_LookupModule.INVESTMENT)

_data_point_lookup_api_request = LookupDataAPIBackend(_LookupModule.DATA_POINT)


class investment_universes:
    All_Managed_Investments = Universe("All Managed Investments", "MI")
    Bonds = Universe("Bonds", "BD")
    Closed_End_Funds = Universe("Closed End Funds", "FC")
    Collective_Investment_Trusts = Universe("Collective Investment Trusts", "CZ")
    European_Pension_Life_Fund_Wrappers = Universe("European Pension/Life Fund Wrappers", "PS")
    eVestment_Hedge_Funds = Universe("eVestment Hedge Funds", "VH")
    eVestment_Separate_Accounts = Universe("eVestment Separate Accounts", "VS")
    Exchange_Traded_Funds = Universe("Exchange-Traded Funds", "FE")
    Funds_Open_End_and_Exchange_Traded_Funds = Universe("Funds (Open End and Exchange-Traded Funds)", "FX")
    France_Dedicated_Funds = Universe("France Dedicated Funds", "DF")
    Global_Restricted_Funds = Universe("Global Restricted Funds", "R1")
    Hedge_Funds = Universe("Hedge Funds", "FH")
    HFR_Hedge_Funds = Universe("HFR Hedge Funds", "H1")
    HSBC_Restricted_Funds = Universe("HSBC Restricted Funds", "IF")
    Insurance_and_Pension_Funds = Universe("Insurance and Pension Funds", "FV")
    Market_Indexes = Universe("Market Indexes", "XI")
    Models = Universe("Models", "MO")
    Money_Market_Funds = Universe("Money Market Funds", "FM")
    Open_End_Funds = Universe("Open End Funds", "FO")
    Plans_529 = Universe("529 Plans", "CP")
    Portfolios_529 = Universe("529 Portfolios", "CT")
    Private_Funds = Universe("Private Funds", "SP")
    Separate_Accounts = Universe("Separate Accounts", "SA")
    Stocks = Universe("Stocks", "ST")
    Strategies = Universe("Strategies", "MG")
    UBS_Separate_Accounts = Universe("UBS Separate Accounts", "S1")
    UK_Life_and_Pension_Funds = Universe("UK Life and Pension Funds", "V1")
    Unit_Investment_Trust = Universe("Unit Investment Trust", "FI")
    US_Variable_Annuities = Universe("US Variable Annuities", "VP")
    US_Variable_Annuity_Subaccounts = Universe("US Variable Annuity Subaccounts", "VA")
    US_Variable_Life = Universe("US Variable Life", "LP")
    US_Variable_Life_Subaccounts = Universe("US Variable Life Subaccounts", "VL")


def _search_data(datapoint: str, keyword: str, universe: str, findBy: str = "name") -> Any:
    # Encode keyword.
    _logger.debug(f"Encode the {keyword} keyword in utf-8")
    keyword_encode = quote(keyword, "utf-8")

    _logger.debug(f"Search by {findBy} in the {datapoint} data point with: keyword: {keyword}, universe: {universe}")
    url = f"{_config.securitydata_service_url()}v1/find?datapoint={datapoint}&kw={keyword_encode}&operator=namecontain&universe={universe}&findby={findBy}&status=activeonly"

    response_json = _investment_lookup_api_request.do_get_request(url)
    return response_json


@_decorator.typechecked
def get_morningstar_data_sets(universe: Optional[str] = None) -> DataFrame:
    """Returns all Morningstar pre-defined data sets.

    Args:
        universe (:obj:`str`, `optional`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.

    :Returns:
        DataFrame: A DataFrame object with Morningstar data sets. DataFrame columns include:

        * datasetId
        * name

    :Examples:
        Retrieve the Morningstar data set for the open-end fund universe.

    ::

        import morningstar_data as md

        df = md.direct.lookup.get_morningstar_data_sets(universe="FO")
        df

    :Output:
        =========  ===============================
        datasetId  name
        =========  ===============================
        0026-0020  Snapshot
        0026-0447  Sustainability: ESG Risk (Fund)
        ...
        =========  ===============================

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    if universe:
        url = f"{_config.data_point_service_url()}v1/defaultviews?type=screen&universe={universe}"
    else:
        url = f"{_config.data_point_service_url()}v1/defaultviews?type=list"
    response_json = _data_set_lookup_api_request.do_get_request(url)
    result = DataFrame(response_json)
    result["datasetId"] = result["id"]
    return result[["datasetId", "name"]]


def _get_multiple_data_point_details(params: list) -> list:
    asset_flow_data_points = []
    normal_data_points = []
    params = params if params else []
    for data_point in params:
        if data_point is None:
            continue
        if data_point.get("datapointId", "").strip() in ALL_ASSET_FLOW_DATA_POINTS:
            asset_flow_data_points.append(data_point.get("datapointId", "").strip())
        else:
            normal_data_points.append(data_point)

    result = []
    if normal_data_points:
        normal_response_json = _get_data_point_details(normal_data_points)
        if isinstance(normal_response_json, list) and len(normal_response_json) > 0:
            result.extend(normal_response_json)

    if asset_flow_data_points:
        result.extend(_get_asset_flow_data_points_by_ids(asset_flow_data_points))

    return _data_points_reorder(params, result)


def _data_points_reorder(params: list, result: list) -> list:
    target = []
    for data_point in params:
        target.extend(_data_points_mapping(data_point, result))
    return target


def _data_points_mapping(data_point: Optional[dict], result: list) -> List[Any]:
    mapping_result: List[Any] = []
    if data_point is None:
        return mapping_result
    data_point_id = data_point.get("datapointId", "").strip()
    is_ts_dp = data_point.get("isTsdp", None)
    if data_point_id in ALL_ASSET_FLOW_DATA_POINTS:
        target_list = list(filter(lambda x: (data_point_id == x.get("datapointId", "").strip()), result))
        if len(target_list) > 0:
            mapping_result.append(target_list[0])
    elif is_ts_dp is not None:
        target_list = list(
            filter(
                lambda x: (data_point_id == x.get("datapointId", "").strip()) & (is_ts_dp == x.get("isTsdp", None)),
                result,
            )
        )
        if len(target_list) > 0:
            mapping_result.append(target_list[0])
    else:
        target_list_for_ts = list(
            filter(
                lambda x: (data_point_id == x.get("datapointId", "").strip()) & (x.get("isTsdp", None) is True),
                result,
            )
        )
        target_list_for_no_ts = list(
            filter(
                lambda x: (data_point_id == x.get("datapointId", "").strip()) & (x.get("isTsdp", None) is False),
                result,
            )
        )
        if len(target_list_for_ts) > 0:
            mapping_result.append(target_list_for_ts[0])
        if len(target_list_for_no_ts) > 0:
            mapping_result.append(target_list_for_no_ts[0])

    return mapping_result


@_decorator.not_null
@_decorator.typechecked
def get_data_point_settings(data_point_ids: List[str]) -> DataFrame:
    """Returns settings for a given set of data points. This settings DataFrame can be manipulated
    to reflect specific settings to be used for data retrieval.

    Args:
        data_point_ids (:obj:`list`): A list of unique identifiers for data points.
            Example: ["OS01W", "HP010"]

    :Returns:
        DataFrame: A DataFrame object with data point settings. DataFrame columns include:

        * datapointId
        * datapointName
        * displayName
        * currency
        * preEuroConversion
        * sourceId
        * frequency
        * startDate
        * endDate
        * floatStart
        * floatEnd
        * startDelay
        * endDelay
        * diffStart
        * diffEnd
        * compounding
        * calculationId
        * annualized
        * annualDays
        * benchmark
        * riskfree
        * windowType
        * windowSize
        * stepSize
        * requireContinueData
        * fit
        * scalType
        * scalValue
        * scalPercentValue
        * timehorizon

    :Examples:
        Get data point settings for data point "OS01W".

    ::

        import morningstar_data as md

        df = md.direct.get_data_point_settings(data_point_ids=["OS01W"])
        df

    :Output:
        ===========  =============  ===  ========  ========  =========
        datapointId  datapointName  ...  calcMnav  showType  transType
        ===========  =============  ===  ========  ========  =========
        OS01W	     Name           ...  None      None      None
        ===========  =============  ===  ========  ========  =========

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    params = list(map(lambda x: {"datapointId": x}, data_point_ids))
    response_json = _get_multiple_data_point_details(params)
    if len(response_json) == 0:
        raise ValueErrorException("There are no data points.")
    extracted_data = _extract_data(response_json)
    details = DataFrame(extracted_data)
    filtered_details = _filter_data_frame_column_by_setting(details)
    details_with_relocated_data_point_id = _reindex_data_frame_column(filtered_details, target_column="datapointId", new_index=0)
    details_with_relocated_data_point_name = _reindex_data_frame_column(
        details_with_relocated_data_point_id, target_column="datapointName", new_index=1
    )
    return _empty_to_none(details_with_relocated_data_point_name)


@_decorator.not_null
@_decorator.typechecked
def investments(
    keyword: str = "", investment: Optional[InvestmentIdentifier] = None, count: int = 50, only_surviving: bool = True
) -> DataFrame:
    """Returns investments that match the given keyword or :class:`~morningstar_data.direct.InvestmentIdentifier` object.

    Args:
        keyword (:obj:`str`, `optional`): Keyword to search for investments.
        investment(:obj:`InvestmentIdentifier`, `optional`): Identifies an investment by its ISIN, CUSIP, ticker, base currency, and/or exchange.
        count (:obj:`int`, `optional`): Maximum number of matching investments to return.
        only_surviving (:obj:`bool`, `optional`): Include only surviving investments.

    .. note::
        You must specify one of, but not both of ``keyword`` or ``investment``.

    :Returns:
        DataFrame: A DataFrame containing the matching investments with the following columns:


        * Name
        * SecId: Morningstar's unique identifier for this investment
        * Ticker
        * ISIN
        * CUSIP
        * Base Currency
        * Exchange: The exchange where the investment is listed
        * Country: The country where the investment is based
        * Security Type: Morningstar security type code for this investment
        * Fund Id
        * Performance Id


    :Examples:

    **1. Find all global listings for a security by its ticker symbol**

    .. code:: python

        import morningstar_data as md
        from morningstar_data.direct import InvestmentIdentifier

        # Define an investment identifier
        investment = InvestmentIdentifier(ticker="AAPL")

        # Retrieve matching investments
        df = md.direct.lookup.investments(investment=investment)
        df

    ==========  ==========  ======  ============  =========  ==============  ========  =======  =============  ========  ===========
    Name        SecId       Ticker  ISIN          CUSIP      Base Currency   Exchange  Country  Security Type  Fund Id   Performance Id
    ==========  ==========  ======  ============  =========  ==============  ========  =======  =============  ========  ===========
    Apple Inc   0P000000GY  AAPL    US0378331005  037833100  USD             EXXNAS    USA      ST             None      0P000000GY
    Apple Inc   0P0001367D  AAPL    US0378331005  037833100  USD             EXXSGO    USA      ST             None      0P0001367D
    ...         ...         ...     ...           ...        ...             ...       ...      ...            ...       ...
    Apple Inc   0P0001KOM0  AAPL    US0378331005  037833100  USD             EXXLIM    USA      ST             None      0P0001KOM0
    Apple Inc   0P0000EEPX  AAPL    US0378331005  037833100  EUR             EXXWBO    USA      ST             None      0P0000EEPX
    ==========  ==========  ======  ============  =========  ==============  ========  =======  =============  ========  ===========

    **2. Get listings using ticker symbol and base currency**

    .. code:: python

        import morningstar_data as md
        from morningstar_data.direct import InvestmentIdentifier

        # Define an investment identifier
        investment = InvestmentIdentifier(ticker="AAPL", base_currency="Canadian Dollar")

        # Retrieve matching investments
        df = md.direct.lookup.investments(investment=investment)
        df

    ================================================== === ============= ======== ======= === ==============
    Name                                               ... Base Currency Exchange Country ... Performance Id
    ================================================== === ============= ======== ======= === ==============
    Apple Inc Canadian Depository Receipt (CAD Hedged) ... CAD           EXXTSE   USA     ... 0P0001N6JJ
    Apple Inc                                          ... CAD           EXXNAS   USA     ... 0P000000GY
    ================================================== === ============= ======== ======= === ==============

    **3. Get listings using ISIN**

    .. code:: python

        import morningstar_data as md
        from morningstar_data.direct import InvestmentIdentifier

        # Define an investment identifier
        investment = InvestmentIdentifier(isin="US5949181045")

        # Retrieve matching investments
        df = md.direct.lookup.investments(investment=investment)
        df

    ============== ========== ====== ============ === =============
    Name           SecId      Ticker ISIN         ... PerformanceId
    ============== ========== ====== ============ === =============
    Microsoft Corp 0P000003MH MSFT   US5949181045 ... 0P000003MH
    Microsoft Corp 0P0000BNJR MSF    US5949181045 ... 0P0000BNJR
    Microsoft Corp 0P0000EGEI MSF    US5949181045 ... 0P0000EGEI
    Microsoft Corp 0P0001AJSV MSFT   US5949181045 ... 0P0001AJSV
    ============== ========== ====== ============ === =============
    """
    columns = [
        "Name",
        "SecId",
        "Ticker",
        "ISIN",
        "CUSIP",
        "Base Currency",
        "Exchange",
        "Country",
        "Security Type",
        "Fund Id",
        "Performance Id",
    ]

    if keyword != "" and investment is not None:
        raise ValueError("Only one of keyword or investment can be specified.")

    if count < 1:
        raise ValueError("Count must be a positive integer.")

    if keyword == "" and investment is not None:
        search_result = mdapi.search_security(investment, count, only_surviving)
        df = DataFrame(search_result.investments)
        return df[columns]

    # Encode keyword.
    keyword_encode = quote(keyword, "utf-8")
    _logger.debug(f"Search investments with:{keyword_encode}, Top {count} of investments to return.")
    url = f"{_config.securitydata_service_url()}v1/autocomplete?checkEntitlement=true&count={count}&keyword={keyword_encode}&onlySurviving={str(only_surviving).lower()}"
    response_json = _investment_lookup_api_request.do_get_request(url)
    investments = response_json["investments"]

    _logger.info(f"No investments return with keyword:{keyword}")
    if not investments:
        _logger.debug(f"No investments return with keyword:{keyword}")
        return DataFrame({"Name": [], "SecId": [], "ISIN": [], "CUSIP": [], "Base Currency": []})

    result = DataFrame(investments)
    # Remove some unused data.
    _logger.info("Remove some unused data")
    _logger.debug(f"{len(investments)} investments return. total investments found:{response_json['total']}")
    new_data_frame = result.drop(columns=["targetUniverseId", "csdccCode", "score", "itaCode"])
    # Rename and reoreder columns.
    _logger.info("Rename and reoreder columns. data")
    new_data_frame = new_data_frame.rename(
        columns={
            "name": "Name",
            "secId": "SecId",
            "ticker": "Ticker",
            "isin": "ISIN",
            "baseCurrency": "Base Currency",
            "cusip": "CUSIP",
            "exchange": "Exchange",
            "countryOrientation": "Country",
            "securityType": "Security Type",
            "fundId": "Fund Id",
            "performanceId": "Performance Id",
        }
    )
    return new_data_frame[columns]


@_decorator.not_null
@_decorator.typechecked
def get_morningstar_category(universe: str) -> DataFrame:
    """Returns Morningstar Categories for the given universe.

    Args:
        universe (:obj:`str`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object with Morningstar Category details. DataFrame columns include:

        * Morningstar Category
        * Category Code
        * Global Broad Category Id
        * Global Broad Category Group
        * Region

    Examples:
        Retrieve the Morningstar Category for the open-end fund universe.

    ::

        import morningstar_data as md

        df = md.direct.lookup.get_morningstar_category("FO")
        df

    :Output:
        ==========================  =============  ========================  ===========================  ==========
        Morningstar Category        Category Code  Global Broad Category Id  Global Broad Category Group  Region
        ==========================  =============  ========================  ===========================  ==========
        10 yr Government Bond	    INCA000008	   $BCG$FXINC	             Fixed Income	              India
        2025 Target Date Portfolio	CACA000158	   $BCG$ALLOC	             Allocation	                  Canada
        2030 Target Date Portfolio	CACA000164	   $BCG$ALLOC	             Allocation	                  Canada
        ...
        ==========================  =============  ========================  ===========================  ==========
    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    url = f"{_config.data_point_service_url()}v1/datapoints/datachoiceoptions?universe={universe}"

    _logger.debug(f"Lookup Morningstar Category from DataPoint API for universe: {universe}")

    # OF003 is datapoint 'Mornignstar Category' which identifies funds based on their actual investment styles as measured by their underlying portfolio holdings
    res = _data_point_lookup_api_request.do_post_request(url, "OF003")

    if not res or not res[0].get("options", []):
        return DataFrame(
            columns=["Morningstar Category", "Category Code", "Global Broad Category Id", "Global Broad Category Group", "Region"]
        )

    options = res[0].get("options", [])

    data = [
        {
            "Morningstar Category": x.get("name", ""),
            "Category Code": x.get("value", ""),
            "Global Broad Category Id": x.get("additional", {}).get("globalBroadCategoryId", ""),
            "Global Broad Category Group": x.get("additional", {}).get("globalBroadCategoryName", ""),
            "Region": x.get("additional", {}).get("regionName", ""),
        }
        for x in options
    ]

    return DataFrame(data)


@_decorator.not_null
@_decorator.typechecked
def firms(keyword: str, universe: Optional[str] = None) -> DataFrame:
    """Returns firms that match the given keyword.

    Args:
        keyword (:obj:`str`): Keyword to search for firms.
        universe (:obj:`str`, `optional`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.

    :Returns:
        DataFrame: A DataFrame object with firms. DataFrame columns include:

        * Id
        * Firm Name
        * Universe


    :Examples:
        Get all firms that match the keyword "rock".

    ::

        import morningstar_data as md

        df = md.direct.lookup.firms("rock")
        df

    :Output:
        =============  =======================================  =============
        Id              Firm Name                                  Universe
        =============  =======================================  =============
        F00000Z17E      Amrock Capital B.V.                             FO
        F00000XC6D      BlackRock	                                    FO
        F00000SEJ6      BlackRock (Channel Islands) Limited	            FO
        ...
        =============  =======================================  =============

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    # Encode keyword.
    keyword_encode = quote(keyword, "utf-8")
    # Encode universe.
    if universe:
        univ_encode = quote(universe.upper(), "utf-8")
    else:
        univ_encode = ""
    _logger.debug(f"Search firms with:{keyword} and universe:{universe} ")
    # OF006 is Datapoint 'Firm Name', A company which offers mutual funds.
    url = f"{_config.securitydata_service_url()}v1/find?datapoint=OF006&findby=namecontain&kw={keyword_encode}&universe={univ_encode}"

    response_json = _investment_lookup_api_request.do_get_request(url)

    firms_df = pandas.json_normalize(response_json)
    if firms_df.empty:
        _logger.debug(f"Search firms without result. keyword:{keyword} and universe:{universe} ")
        return DataFrame({"Id": [], "Name": [], "Universe": []})
    firms_df = firms_df.rename(columns={"name": "Name", "additional.universe": "Universe", "id": "Id"})
    return firms_df


@_decorator.typechecked
def get_brandings(keyword: str, find_by: str, universe: Optional[str] = None) -> DataFrame:
    """Returns brandings for the given universe and keyword.

    Args:
        keyword(:obj:`str`): Keyword to search for brandings.
        find_by(:obj:`str`): Search condition to search for brandings. Valid values are ``name_begin`` and ``name_contain``.
        universe(:obj:`str`, `optional`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.

    :Returns:
        DataFrame: A DataFrame object with brandings. DataFrame columns include:

        * Id
        * Name

    :Examples:
        Search brandings in the open-end fund universe, where branding name begins with the keyword "a".

    ::

        import morningstar_data as md

        df = direct.lookup.get_brandings("a", "name_begin", "FO")
        df

    :Output:
        ==========  ===============================
        Id            Name
        ==========  ===============================
        BN000007Q5    A Plus Finance
        BN000007Q6    A&G
        BN00000GBI    A1
        ...
        ==========  ===============================

    :Errors:
        BadRequestException: Raised when the parameter find_by is not a valid value.

        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.
    """

    if find_by not in ["name_begin", "name_contain"]:
        raise BadRequestException(message="Valid values for find_by should be `name_begin` or `name_contain`.")

    condition = find_by.replace("_", "")

    # FC001 is a global grouping of asset management companies that represents the management philosophy,
    # the firm's marketing/communication channels, and their distribution/sales efforts.
    url = f"{_config.securitydata_service_url()}v1/find?datapoint=FC001&findby={condition}&kw={keyword}&universe={universe}"

    res = _investment_lookup_api_request.do_get_request(url)

    if not res:
        return DataFrame(columns=["Id", "Name"])

    data = [{"Id": x.get("id", ""), "Name": x.get("name", "")} for x in res]

    return DataFrame(data)


@_decorator.typechecked
def portfolio_managers(keyword: str = "", universe: str = "FO") -> DataFrame:
    """
    Returns portfolio managers for the given universe and keyword.

    Args:
        keyword (:obj:`str`): String to match one or more manager names. If the keyword is empty or not passed to the function, the function will return all managers.

        universe (:obj:`str`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.


    :Returns:
        DataFrame: A DataFrame object with portfolio managers. DataFrame columns include:

        * Id
        * Manager Name
        * Firm

    :Examples:
        Get portfolio managers that match the keyword "alla".

    ::

        import morningstar_data as md

        df = md.direct.lookup.portfolio_managers(keyword="alla")
        df

    :Output:
        ======  =======================  ===============
        Id   	Manager name	         Firm
        ======  =======================  ===============
        81643	Abdallah Guezour	     Schroders
        123653	Abdallah Nauphal	     TD
        213243	Adam Rizkalla	         River Canyon
        153647	Adrian Allardice	     Old Mutual
        182699	Adrian van Pallander	 Coronation
        ...
        ======  =======================  ===============

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    # FCM01 is Datapoint which name is 'Manager & Fund Company'
    _logger.debug(f"Search managers with: keyword: {keyword}, universe: {universe} in the FCM01 datapoint")
    data = _search_data("FCM01", keyword, universe)

    _logger.info(f"Total managers found: {len(data)} managers")
    result = DataFrame(data)

    _logger.debug("Rename columns")
    result.rename(columns={"id": "Id", "name": "Manager name"}, inplace=True)

    if result.empty:
        _logger.info(f"No managers return with keyword: {keyword}")
        return DataFrame({"Id": [], "Manager name": [], "Firm": []})

    _logger.debug("Select only the Id and Manager name in the DataFrame")
    portfolio_managers_result = result[["Id", "Manager name"]]

    _logger.debug("Apply the _separate_name_and_firm method in the columns to separate the manager name and the firm")
    portfolio_managers_result = portfolio_managers_result.apply(_separate_name_and_firm, axis=1)

    return portfolio_managers_result


# The name column has the manager's name and the firm name in the same column. This function looks for separating them into different columns
# INPUT: name column: "Abdallah Guezour (Schroders)"
# OUTPUT: name column: "Abdallah Guezour | firm column: "Schroders"
def _separate_name_and_firm(portfolio_manager: DataFrame) -> DataFrame:
    manager_name = portfolio_manager["Manager name"]
    # Check if the name does not contain the firm within (). If don't, the firm value will be None
    if manager_name.count("(") < 1:
        portfolio_manager["Firm"] = None
    # If the name contains the firm name within "()" we will separate them in different columns
    else:
        # Extract the firm name with an regex expression based on the name value
        portfolio_manager["Firm"] = re.findall("\((.+?)\)", manager_name)[0]
        # Remove signature name from column name value
        portfolio_manager["Manager name"] = manager_name.replace("(" + portfolio_manager["Firm"] + ")", "").strip()

    return portfolio_manager


@_decorator.typechecked
def companies(keyword: str = "", universe: str = "ST") -> DataFrame:
    """
    Returns companies for the given universe and keyword.

    Args:
        keyword (:obj:`str`): String to match one or more company names. If the keyword is empty or not passed to the function, the function will return all companies.

        universe (:obj:`str`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.


    :Returns:
        DataFrame: A DataFrame object with companies. DataFrame columns include:

        * Id
        * Company name

    :Examples:
        Get companies that match the keyword "Energy".

    ::

        import morningstar_data as md

        df = md.direct.lookup.companies("Energy")
        df

    :Output:
        ==========  ========================
        Id	        Company name
        ==========  ========================
        0C00008O2A	11 Good Energy Inc
        0C000017RQ	2G Energy AG
        0C0000516I	3MV Energy Corp
        0C00000ZAZ	3Power Energy Group Inc
        0C00000QLB	3TEC Energy Corp
        ...
        ==========  ========================

    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    # AA0A5 is Datapoint which name is 'Company Name'
    _logger.debug(f"Search companies with: keyword: {keyword}, universe: {universe} in the AA0A5 datapoint")
    data = _search_data("AA0A5", keyword, universe)

    _logger.info(f"Total companies found: {len(data)} companies")
    result = DataFrame(data)

    _logger.debug("Rename columns")
    result.rename(columns={"id": "Id", "name": "Company name"}, inplace=True)

    if result.empty:
        _logger.info(f"No companies return with keyword: {keyword}")
        return DataFrame({"Id": [], "Company name": []})

    _logger.debug("Select only the Id and Company name in the DataFrame")
    companies_result = result[["Id", "Company name"]]

    return companies_result


@_decorator.typechecked
def get_investment_universes(include_category_universes: bool = False) -> DataFrame:
    """Returns investment universe names and IDs.
    For example, Bonds, eVestment Hedge Funds, Global Restricted Funds.

    Args:
        include_category_universes (:obj:`bool`, `optional`): If True, the function will return all universes, including category universes. Default is False.

    :Returns:
        DataFrame: A DataFrame object with universes.  DataFrame columns include:

        * Id
        * Name

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.lookup.get_investment_universes()
        df

    :Output:
        ==== ==============
        Id   Name
        ==== ==============
        FH   Hedge Funds
        XI   Market_Indexes
        ...
        ==== ==============
    """

    universes = _get_all_universes()

    universes = universes.rename(columns={"id": "Id", "name": "Name"})

    if not include_category_universes:
        universes = universes[~universes["Id"].str.contains("CA]")]  # Drops all category universes

    return universes



from ._exceptions import (
    AccessDeniedError,
    BadRequestError,
    MdApiRequestException,
    MdApiTaskException,
    MdApiTaskFailure,
    MdApiTaskTimeoutException,
    MdBaseException,
    ResourceNotFoundError,
    exception_by_name,
)
from ._mdapi import call_remote_function, search_security
from ._types import (
    InvestmentLookupResult,
    MdapiTask,
    RequestObject,
    TaskResult,
    TaskStatus,
)

__all__ = [
    "MdBaseException",
    "MdApiRequestException",
    "MdApiTaskException",
    "MdApiTaskFailure",
    "MdApiTaskTimeoutException",
    "AccessDeniedError",
    "BadRequestError",
    "ResourceNotFoundError",
    "exception_by_name",
    "RequestObject",
    "MdapiTask",
    "TaskResult",
    "TaskStatus",
    "call_remote_function",
    "search_security",
    "InvestmentLookupResult",
]



import warnings
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from pandas import DataFrame

from .. import mdapi
from .._utils import (
    _get_data_points_total_columns,
    _get_user_cells_quota,
    data_point_dataframe_to_list,
)
from ..mdapi import RequestObject
from . import _decorator
from ._data_objects import DataPoints, Investments
from ._data_type import DryRunResults
from .data_type import InvestmentIdentifier, TimeSeriesFormat, WarningBehavior

NUMBER_OF_INVESTMENTIDENTIFIERS_ALLOWED = 500


@dataclass
class InvestmentDataRequest(RequestObject):
    investments: Union[List[str], str, Dict[str, Any], List["InvestmentIdentifier"]]
    datapoints: Optional[Union[List[Dict[str, Any]], str, List[Any]]] = None
    time_series_format: TimeSeriesFormat = TimeSeriesFormat.WIDE
    display_name: Optional[bool] = False
    unmatched_investment_behavior: Optional[WarningBehavior] = WarningBehavior.WARNING
    preview: Optional[bool] = False

    def check_size_of_request(self) -> None:
        if isinstance(self.investments, list) and len(self.investments) > NUMBER_OF_INVESTMENTIDENTIFIERS_ALLOWED:
            if all(isinstance(item, InvestmentIdentifier) for item in self.investments):
                raise ValueError(
                    f"The size of investments list should not be greater than {NUMBER_OF_INVESTMENTIDENTIFIERS_ALLOWED} if of type List[InvestmentIdentifier]"
                )


@_decorator.typechecked
def investment_data(
    investments: Union[List[str], str, Dict[str, Any]],
    datapoints: Union[List[Dict[str, Any]], str, DataFrame],
) -> DataFrame:
    warnings.warn(
        "The investment_data function is deprecated and will be removed in the next major version. Use get_investment_data instead",
        FutureWarning,
        stacklevel=2,
    )
    return _get_investment_data(investments, datapoints, time_series_format=TimeSeriesFormat.WIDE)


# FIXME: In https://msjira.morningstar.com/browse/AL2-92, the error_handler decorator was removed from some methods
# to trigger QueryLimitException.
# There are multiple methods where the decorator was removed to trigger certain exceptions back to the caller:
#   - save_investment_list
#   - investment_data
#   - holding_dates
#   - asset_flow
# For the time being, we will leave this as-is to keep things working. Moving forward, we must re-design the
# exception handling so that we don't have these types of special cases where error_handler is only allowed on some
# methods but not others.
@_decorator.typechecked
def get_investment_data(
    investments: Union[List[str], str, Dict[str, Any], List[InvestmentIdentifier]],
    data_points: Optional[Union[List[Dict[str, Any]], str, DataFrame, List[Any]]] = None,
    display_name: bool = False,
    dry_run: Optional[bool] = False,
    time_series_format: TimeSeriesFormat = TimeSeriesFormat.WIDE,
    unmatched_investment_behavior: Optional[WarningBehavior] = WarningBehavior.WARNING,
    preview: Optional[bool] = False,
) -> Union[DataFrame, DryRunResults]:
    """Retrieve data for the specified investments and data points.

    Args:
        investments (:obj:`Union`, `required`): Defines the investments to fetch. Input can be:

            * Investment IDs (:obj:`list`, `optional`): Investment identifiers, in the format of SecId;Universe or just SecId. E.g., ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"]. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
            * InvestmentIdentifiers (:obj:`list`, `optional`): A list of :class:`~morningstar_data.direct.InvestmentIdentifier` objects, allowing users to specify investments using standard identifiers such as ISIN, CUSIP, and/or ticker symbol instead of Morningstar SecIds. Introduced in version 1.11.0.

              * Multiple Matches: If multiple valid matches exist for a single InvestmentIdentifier, results will be prioritized according to the :ref:`security matching logic<Security Matching Logic>`. The highest ranked security will be used.
              * Request Limit: Supports up to 500 InvestmentIdentifier objects per request.

            * Investment List ID (:obj:`str`, `optional`): Saved investment list in Morningstar Direct. Currently, this function does not support lists that combine investments and user-created portfolios. Use the `get_investment_lists <./lists.html#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.
            * Search Criteria  ID (:obj:`str`, `optional`): Saved search criteria in Morningstar Direct. Use the `get_search_criteria <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria>`_ function to discover saved search criteria.
            * Search Criteria Condition (:obj:`dict`, `optional`): Search criteria definition. See details in the Reference section below or use the `get_search_criteria_conditions <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria_conditions>`_ function to discover the definition of a saved search criteria.

        data_points (:obj:`Union`, `optional`): Defines the data points to fetch. If not provided and investments are specified with a list ID or search criteria ID, the corresponding bound dataset will be used.

            * Data Point IDs (:obj:`List[Dict]`, `optional`): A list of dictionaries, each defining a data point and its (optional) associated settings.
            * Data Set ID (:obj:`str`, `optional`): Morningstar data set or user-created data set saved in Morningstar Direct. Use the `get_data_sets <./data_set.html#morningstar_data.direct.user_items.get_data_sets>`_ or `get_morningstar_data_sets <./data_set.html#morningstar_data.direct.get_morningstar_data_sets>`_ functions to discover saved data sets.
            * Data Point Settings (:obj:`DataFrame`, `optional`): A DataFrame of data point identifiers and their associated settings. Use the `get_data_set_details <./data_set.html#morningstar_data.direct.user_items.get_data_set_details>`_ function to discover data point settings from a saved data set.

        display_name (:obj:`bool`, `optional`): When true, the returned column names will match display names saved in the data set. Default is false.
        time_series_format (:obj:`direct.TimeSeriesFormat`, `optional`): Specifies the format of the time series data. Default is WIDE. Accepted values are:

            * ``TimeSeriesFormat.WIDE``: Data is presented in a wide format, where each row contains a different investment with values for each point in time in a different column.
            * ``TimeSeriesFormat.LONG``: Data is presented in a long format, where each row represents a single observation for a variable at a point in time.
            * ``TimeSeriesFormat.LONG_WITHOUT_NAME``: Similar to the long format but excludes the 'Name' column.

        dry_run(:obj:`bool`, `optional`): When True, the query will not be executed. Instead, a DryRunResults object will be returned with details about the query's impact on daily cell limit usage.
        unmatched_investment_behavior (:obj:`WarningBehavior`, `optional`): Determines behavior when any of the input investments can't be found.

            * ``WarningBehavior.FAIL``: A `ResourceNotFoundError` is raised if any of the investments can't be found.
            * ``WarningBehavior.WARNING`` (default): A warning is displayed listing any investments that were not found. A `ResourceNotFoundError` is raised if all investments were not found.
            * ``WarningBehavior.IGNORE``: No warning is displayed if any investments were not found. A `ResourceNotFoundError` is raised if all investments were not found.

        preview (:obj:`bool`, `optional`): Defaults to False. Setting to True allows access to data points outside of your current subscription, but limits the output to 25 rows.


    :Returns:

        There are two return types:

        * DataFrame: A DataFrame object with investment data

        * DryRunResults: Is returned if dry_run=True is passed

          * estimated_cells_used: Number of cells by this query
          * daily_cells_remaining_before: How many cells are remaining in your daily cell limit before running this query
          * daily_cells_remaining_after: How many cells would be remaining in your daily cell limit after running this query
          * daily_cell_limit: Your total daily cell limit

    :Reference:
        Constructing a Search Criteria Condition dictionary:

        For example::

                    SEARCH_CRITERIA_CONDITION = {
                            "universeId": "cz",
                            "subUniverseId": "",
                            "subUniverseName": "",
                            "securityStatus": "activeonly",
                            "useDefinedPrimary": False,
                            "criteria": [
                                {"relation": "", "field": "HU338", "operator": "=", "value": "1"},
                                {"relation": "AND", "field": "HU863", "operator": "=", "value": "1"}
                            ]
                        }

        * universeId (:obj:`string`, `required`): Universe code (e.g., "FO"). Use the `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ function to explore available values or `download <../_static/assets/SearchCriteriaInfo.xlsx>`_ the reference file.
        * subUniverseId (:obj:`string`, `optional`): Sub-universe code, if applicable. See the refererece file above for available values.
        * subUniverseName (:obj:`string`, `optional`): Name of sub-universe, if applicable. See the refererece file above for available values.
        * securityStatus (:obj:`string`, `optional`): Security status, can be 'activeonly' or 'activeinactive'.
        * useDefinedPrimary (:obj:`Boolean`, `optional`): If set to true, Morningstar Direct user-defined settings will be used.
        * criteria (:obj:`List[Dict[]]`, `required`): Custom search conditions. Dictionary fields described below.


          * field (:obj:`string`): Data point identifier
          * value (:obj:`string`): Value to compare against the data point.
          * criteria (:obj:`List[Dict[]]`): A nested list of additional custom search conditions.
          * relation (:obj:`string`, `required`): Boolean condition used when joining with the previous condition. Accepts an empty string (for the first condition, as no previous condition exists to join with), 'Or', or 'And'.
          * operator (:obj:`string`): Operator used to compare field value to value. Possible operators include '=', '!=', '<', '>', '<=', '>=', 'like' (data contains value), and 'not like' (data does not contain value).

            For example::

                [
                    {
                        "field": "eb919bcc-c097-4fe3-898c-470d8b89dde1"
                        "operator": "="
                        "relation": ""
                        "value": "122"
                    },
                    {
                        "criteria": [
                            {...Condition1 },
                            {...Condition2 },
                            {...Condition3 },
                        ],
                        "relation": "And"
                    },
                    {
                        "field": "LS466"
                        "operator": "="
                        "relation": "Or"
                        "value": "FH"
                    }
                ]

    ----------------------
    Usage Examples
    ----------------------

    **Get investment data for the given investments and data points.**
    ::

        import morningstar_data as md

        df = md.direct.get_investment_data(
            investments=["F0AUS05U7H", "F000010NJ5"],
            data_points=[
                {"datapointId": "OS01W"}, # Name
                {"datapointId": "LS05M"}, # Base Currency
                {
                    "datapointId": "HP010", # Monthly Return
                    "isTsdp": True,
                    "currency": "CNY",
                    "startDate": "2021-03-01",
                    "endDate": "2021-08-31",
                },
            ],
        )
        df

    :Output:
        =================  ===================================  ================= =========================  =========================
        Id                 Name                                 Base Currency     Monthly Return 2021-03-31  Monthly Return 2021-04-30
        =================  ===================================  ================= =========================  =========================
        F0AUS05U7H         Walter Scott Global Equity           Australian Dollar                  3.726094                   3.078352
        F000010NJ5         Vontobel Emerging Markets Eq U1 USD  US Dollar                         -0.417526                  -0.376890
        =================  ===================================  ================= =========================  =========================


    **Get investment data for a saved investment list and data points.**

    ::

        import morningstar_data as md

        df = md.direct.get_investment_data(
            investments="a727113a-9557-4378-924f-5d2ba553f687", # Replace with a valid List ID
            data_points=[{"datapointId": "HS793", "isTsdp": True}],
        )
        df


    ==============  =================================  =============================  =============================  =============================
    Id              Name                               Daily Return Index 2021-09-23  Daily Return Index 2021-09-24  Daily Return Index 2021-09-25
    ==============  =================================  =============================  =============================  =============================
    FOUSA00DFS;FO   BlackRock Global Allocation Inv A  129.92672                      129.56781                      129.56781
    ==============  =================================  =============================  =============================  =============================

    **Get investment data for a saved search criteria and data points.**

    ::

        import morningstar_data as md

        df = md.direct.get_investment_data(
            investments="4216254", # Replace with a valid Search Criteria ID
            data_points=[{"datapointId": "12", "isTsdp": True}]
        )
        df


    ==============  =======================  =============================
    Id              Name                     Beta 2018-10-01 to 2021-09-30
    ==============  =======================  =============================
    FOUSA06JNH;FO   DWS RREEF Real Assets A    0.654343
    ==============  =======================  =============================

    **Get investment data for a custom search criteria and data points.**

    ::

        import morningstar_data as md

        SEARCH_CRITERIA_CONDITION = {
            "universeId": "cz",
            "subUniverseId": "",
            "subUniverseName": "",
            "securityStatus": "activeonly",
            "useDefinedPrimary": False,
            "criteria": [
                {"relation": "", "field": "HU338", "operator": "=", "value": "1"},
                {"relation": "AND", "field": "HU863", "operator": "=", "value": "1"},
            ],
        }

        df = md.direct.get_investment_data(
            investments=SEARCH_CRITERIA_CONDITION,
            data_points=[{"datapointId": "HS793", "isTsdp": True}],
        )
        df


    =  ===============  =====================================  =============================  ===  =============================
    #  Id               Name                                   Daily Return Index 2022-02-18  ...  Daily Return Index 2022-03-17
    =  ===============  =====================================  =============================  ===  =============================
    0  FOUSA06UOR;CZ    Columbia Trust Stable Government Fund  None                           ...  None
    1  FOUSA06UWL;CZ    Columbia Trust Stable Income Fund      88.8333                        ...  90.7781
    =  ===============  =====================================  =============================  ===  =============================

    **Get investment data for the given investments and data points in the long format.**
    ::

        import morningstar_data as md

        df = md.direct.get_investment_data(
            investments=["F0AUS05U7H", "F000010NJ5"],
            data_points=[
                {"datapointId": "OS01W"}, # Name
                {"datapointId": "LS05M"}, # Base Currency
                {
                    "datapointId": "HP010", # Monthly Return
                    "isTsdp": True,
                    "currency": "CNY",
                    "startDate": "2021-03-01",
                    "endDate": "2021-08-31",
                },
            ],
            time_series_format=md.direct.data_type.TimeSeriesFormat.LONG
        )
        df


    =================  ===================================  ================= =========================  =========================
    Id                 Name                                 Base Currency     Date                       Monthly Return
    =================  ===================================  ================= =========================  =========================
    F0AUS05U7H         Walter Scott Global Equity           Australian Dollar                2021-04-30                   3.726094
    F0AUS05U7H         Walter Scott Global Equity           Australian Dollar                2021-05-31                   3.078352
    F0AUS05U7H         Walter Scott Global Equity           Australian Dollar                2021-06-30                   3.078352
    F000010NJ5         Vontobel Emerging Markets Eq U1 USD  US Dollar                        2021-04-30                  -0.417526
    F000010NJ5         Vontobel Emerging Markets Eq U1 USD  US Dollar                        2021-05-31                  -0.376890
    F000010NJ5         Vontobel Emerging Markets Eq U1 USD  US Dollar                        2021-06-30                  3.078352
    =================  ===================================  ================= =========================  =========================

    **Using Ticker to retrieve data for Apple (AAPL).**
    ::

        import morningstar_data as md
        from morningstar_data.direct import InvestmentIdentifier

        # Define the data point IDs to retrieve
        data_point_ids = [
            {"datapointId": "LS05M"}, # Base Currency
            {"datapointId": "LS01Z"}, # Exchange by Name
            {"datapointId": "OS00F"},  # Inception Date
            {"datapointId": "LS017"}   # Domicile
        ]

       # Initialize an InvestmentIdentifier object for the ticker symbol
       investments = [InvestmentIdentifier(ticker="AAPL")]

       # Retrieve investment data using the defined ticker and data points
       df = md.direct.get_investment_data(investments, data_point_ids)

       df

    ============ =========== =============== ====================== ================ ==============
    Id           Name        Base Currency   Exchange               Inception Date   Domicile
    ============ =========== =============== ====================== ================ ==============
    0P000000GY   Apple Inc   US Dollar       Nasdaq - All Markets   1980-12-12       United States
    ============ =========== =============== ====================== ================ ==============


    **Using Ticker and Currency to retrieve data for Apple (AAPL).**
    ::

        import morningstar_data as md
        from morningstar_data.direct import InvestmentIdentifier

        # Define the data point IDs to retrieve
        data_point_ids = [
            {"datapointId": "LS05M"}, # Base Currency
            {"datapointId": "LS01Z"}, # Exchange by Name
            {"datapointId": "OS00F"},  # Inception Date
            {"datapointId": "LS017"}   # Domicile
        ]

        # Example: Using Ticker and Base Currency
        investments = [InvestmentIdentifier(ticker="AAPL", base_currency="Euro")]

        # Retrieve the data via ticker
        df = md.direct.get_investment_data(investments, data_point_ids)

        df

    ============ =========== =============== ==================== ================ ==============
    Id           Name        Base Currency   Exchange             Inception Date   Domicile
    ============ =========== =============== ==================== ================ ==============
    0P0000EEPX   Apple Inc   Euro            Wiener Boerse AG     2017-05-23       United States
    ============ =========== =============== ==================== ================ ==============


    **Retrieving Investment Data Using Ticker and Exchange.**
    ::

        import morningstar_data as md
        from morningstar_data.direct import InvestmentIdentifier

        # Define the data point IDs to retrieve
        data_point_ids = [
            {"datapointId": "LS05M"}, # Base Currency
            {"datapointId": "LS01Z"}, # Exchange by Name
            {"datapointId": "OS00F"},  # Inception Date
            {"datapointId": "LS017"}   # Domicile
        ]

        # Example: Using Ticker and Exchange
        investments = [InvestmentIdentifier(ticker="MSFT", exchange="Toronto Stock Exchange")]

        # Retrieve the data via Ticker and Exchange
        df = md.direct.get_investment_data(investments, data_point_ids)

        df

    ============ ============================================ ================== ======================== ================ ===============
    Id           Name                                         Base Currency      Exchange                 Inception Date   Domicile
    ============ ============================================ ================== ======================== ================ ===============
    0P0001NHUC   Microsoft Corp Canadian Depository Receipt   Canadian Dollar	   Toronto Stock Exchange   2021-10-05       United States
    ============ ============================================ ================== ======================== ================ ===============

    **Retrieving Investment Data Using ISINs and CUSIPs.**
    ::

        import morningstar_data as md
        from morningstar_data.direct import InvestmentIdentifier

        # Defining Datapoints
        data_point_ids = [
            {"datapointId": "LS05M"},  # Base Currency
            {"datapointId": "LS01Z"},  # Exchange by Name
            {"datapointId": "LF035"},  # Global Broad Category Group
            {"datapointId": "ZZ006"},  # Annual Expense Ratio
        ]

        # Example: Using ISINs and CUSIPs
        investments = [
            InvestmentIdentifier(isin="US6176971074"),  # ISIN for the Morningstar US Equity Fund
            InvestmentIdentifier(cusip="453320103"),    # CUSIP for the Income Fund of America
        ]

        # Retrieve the data via ISIN and CUSIP
        df = md.direct.get_investment_data(investments, data_point_ids)

        df

    === ============= ======================================= =============== ======================= ============================= ======================================
    #   Id            Name                                    Base Currency   Exchange                Global Broad Category Group   Annual Report Adjusted Expense Ratio
    === ============= ======================================= =============== ======================= ============================= ======================================
    1   F00000YOOK    Morningstar US Equity                   US Dollar       Nasdaq - All Markets    Equity                        0.84
    2   FOUSA00D6E    American Funds Income Fund of Amer A    US Dollar       Nasdaq - All Markets    Allocation                    0.58
    === ============= ======================================= =============== ======================= ============================= ======================================


    **Bulk Processing via Identifier Lists.**
    ::

        import morningstar_data as md
        from morningstar_data.direct import InvestmentIdentifier
        import pandas as pd

        # Defining Data Points
        data_point_ids = [
            {"datapointId": "LS05M"},  # Base Currency
            {"datapointId": "LS01Z"},  # Exchange by Name
            {"datapointId": "LF035"},  # Global Broad Category Group
            {"datapointId": "ZZ006"},  # Annual Expense Ratio
        ]

        # Read the list of ISINs from securities.csv
        df_securities = pd.read_csv("securities.csv")

        # Extract the list of ISINs from the DataFrame
        isin_list = df_securities['ISIN'].tolist()

        # Create InvestmentIdentifier objects using the ISINs list
        investments_isins = [InvestmentIdentifier(isin=isin) for isin in isin_list]

        # Retrieve the securities and the associated data_point_ids simply using ISINs
        df = md.direct.get_investment_data(
            investments=investments_isins,
            data_points=data_point_ids
        )

        df

    === ============= ===================================== =============== ======================= ============================= ======================================
    #   Id            Name                                  Base Currency   Exchange                Global Broad Category Group   Annual Report Adjusted Expense Ratio
    === ============= ===================================== =============== ======================= ============================= ======================================
    1   FOUSA00K3V    AB Concentrated Growth Advisor        US Dollar       Nasdaq - All Markets    Equity                        0.75
    2   FOUSA00CAQ    AB Growth A                           US Dollar       Nasdaq - All Markets    Equity                        1.09
    3   FOUSA00B92    AB Large Cap Growth A                 US Dollar       Nasdaq - All Markets    Equity                        0.81
    4   F00000YDG4    AB Sustainable US Thematic Advisor    US Dollar       Nasdaq - All Markets    Equity                        0.65
    5   FOUSA00LBE    abrdn US Sustainable Leaders A        US Dollar       Nasdaq - All Markets    Equity                        1.19
    === ============= ===================================== =============== ======================= ============================= ======================================



    .. Note::
        ``InvestmentSearchWarning:`` This warning is raised when the get_investment_data function is unable to find investments for all provided fields in InvestmentIdentifier objects. If you encounter this warning, review your identifiers to ensure they are valid and correctly specified.

    """

    if len(investments) == 0:
        raise ValueError("Please specify at least one investment")

    result = _get_investment_data(
        investments=investments,
        data_points=data_points,
        display_name=display_name,
        time_series_format=time_series_format,
        dry_run=dry_run,
        unmatched_investment_behavior=unmatched_investment_behavior,
        preview=preview,
    )

    return result


@_decorator.typechecked
def _get_investment_data(
    investments: Union[List[str], str, Dict[str, Any], List[InvestmentIdentifier]],
    data_points: Optional[Union[List[Dict[str, Any]], str, DataFrame, List[Any]]] = None,
    display_name: bool = False,
    dry_run: Optional[bool] = False,
    time_series_format: TimeSeriesFormat = TimeSeriesFormat.WIDE,
    unmatched_investment_behavior: Optional[WarningBehavior] = WarningBehavior.WARNING,
    preview: Optional[bool] = False,
) -> Union[DataFrame, DryRunResults]:
    if dry_run:
        investment_param = Investments(investments)
        data_point_param = DataPoints(data_points)
        investment_id_list = investment_param.get_investment_ids()
        data_point_settings = data_point_param.get_data_points()
        data_point_settings = data_point_settings.where(data_point_settings.notnull(), None)
        data_point_list = data_point_settings.to_dict(orient="records")
        estimated_cells_used = _get_data_points_total_columns(data_point_list) * len(investment_id_list)
        user_cells_quota = _get_user_cells_quota()
        dry_run_results = DryRunResults(
            estimated_cells_used=estimated_cells_used,
            daily_cells_remaining_before=user_cells_quota["daily_cell_remaining"],
            daily_cell_limit=user_cells_quota["daily_cell_limit"],
        )
        return dry_run_results

    if isinstance(data_points, DataFrame):
        data_points = data_point_dataframe_to_list(data_points)

    investment_data_request = InvestmentDataRequest(
        investments=investments,
        datapoints=data_points,
        time_series_format=time_series_format,
        display_name=display_name,
        unmatched_investment_behavior=unmatched_investment_behavior,
        preview=preview,
    )

    investment_data_request.check_size_of_request()

    return mdapi.call_remote_function(
        "get_investment_data",
        investment_data_request,
    )
